#include <utility>

#include <utility>


#include <memory>
#include "Evaluator.hpp"
#include "EvaluationException.hpp"
#include "../Script.hpp"
#include "EvaluationScope/EvaluationScope.hpp"
#include "EvalValues/ScriptFunctionEvalValue.hpp"

using namespace std;

void Evaluator::Evaluate(BoundScriptStatement *statement) {
    this->_evaluationScope = new EvaluationScope(this->_scriptData->_scriptVariables, statement->GetDeepestScope());
    EvaluateBlockStatement(statement);
}

void Evaluator::EvaluateStatement(BoundStatement *statement) {
    switch (statement->GetKind()){
        case BoundStatementKind ::Script: throw; // Should never happen
        case BoundStatementKind ::Block: return this -> EvaluateBlockStatement((BoundBlockStatement*)statement);
        case BoundStatementKind ::Expression: return this -> EvaluateExpressionStatement((BoundExpressionStatement*)statement);
        case BoundStatementKind ::Assignment: return this -> EvaluateAssignmentStatement((BoundAssignmentStatement*)statement);
        case BoundStatementKind ::FunctionDeclaration: return this->EvaluateFunctionDeclarationStatement((BoundFunctionDeclarationStatement*)statement);

        case BoundStatementKind::Bad:
            throw;
    }
}

void Evaluator::EvaluateBlockStatement(BoundBlockStatement* statement) {
    this->_evaluationScope->OuterScope();
    for (auto s: statement->GetStatements()){
        this -> EvaluateStatement(s);
    }
    this->_evaluationScope->InnerScope();
}

void Evaluator::EvaluateExpressionStatement(BoundExpressionStatement *statement) {
    // Save new value
    this->_lastValue = this -> EvaluateExpression(statement->GetExpression());
}

void Evaluator::EvaluateAssignmentStatement(BoundAssignmentStatement *statement) {
    auto value = this -> EvaluateExpression(statement->GetExpression());
    auto key = statement->GetKey();
    if (key->IsCreation()){
        this->_evaluationScope->CreateVariable(key->GetScopeId(), key->GetIdentifier(), value);
    } else{
        this->_evaluationScope->SetVariable(key->GetScopeId(), key->GetIdentifier(), value);
    }
}

void Evaluator::EvaluateFunctionDeclarationStatement(BoundFunctionDeclarationStatement *statement) {
    auto type = statement->GetType();
    auto key = statement->GetKey();
    auto block = statement->GetBlock();
    auto value = make_shared<ScriptFunctionEvalValue>(block, type);
    if (key->IsCreation()){
        this->_evaluationScope->CreateVariable(key->GetScopeId(), key->GetIdentifier(), value);
    } else{
        this->_evaluationScope->SetVariable(key->GetScopeId(), key->GetIdentifier(), value);
    }
}

shared_ptr<EvalValue> Evaluator::EvaluateExpression(BoundExpression *expression) {
    auto type = expression -> GetType();
    switch (type->GetClass()){
        case TypeClass ::Number: return this -> EvaluateIntegerExpression(expression);
        case TypeClass ::Bool: return this -> EvaluateBoolExpression(expression);
        case TypeClass ::String: return this -> EvaluateStringExpression(expression);
        case TypeClass ::Function: return this->EvaluateFunctionExpression(expression);
        case TypeClass ::Nil: return this->EvaluateNilExpression(expression);
        default: throw;
    }
}

shared_ptr<EvalValue> Evaluator::GetVariable(BoundVariableExpression* expression){
    return this->_evaluationScope->GetVariable(expression->GetScope(), expression->GetId())->Clone();
}

shared_ptr<NumericEvalValue> Evaluator::EvaluateIntegerExpression(BoundExpression *expression) {
    switch (expression->GetKind()){
        case BoundExpressionKind ::LiteralInteger: return make_shared<IntegerEvalValue>(((BoundLiteralIntegerExpression*)expression)->GetValue());
        case BoundExpressionKind ::LiteralFloat: return make_shared<FloatEvalValue>(((BoundLiteralFloatExpression*)expression)->GetValue());
        case BoundExpressionKind::Unary: return this -> EvaluateIntegerUnary((BoundUnaryExpression*)expression);
        case BoundExpressionKind ::Binary: return this -> EvaluateIntegerBinary((BoundBinaryExpression*)expression);
        case BoundExpressionKind::Variable: return dynamic_pointer_cast<NumericEvalValue>(this->GetVariable((BoundVariableExpression*)expression));
        case BoundExpressionKind ::FunctionCall: return dynamic_pointer_cast<NumericEvalValue>(this->EvaluateFunctionCallExpression(expression));

        case BoundExpressionKind ::LiteralString:
        case BoundExpressionKind ::LiteralBool:
        case BoundExpressionKind ::Bad:
            throw;
    }
}

shared_ptr<BooleanEvalValue> Evaluator::EvaluateBoolExpression(BoundExpression *expression) {
    switch (expression->GetKind()) {
        case BoundExpressionKind ::LiteralBool: return make_shared<BooleanEvalValue>(((BoundLiteralBoolExpression*)expression)->GetValue());
        case BoundExpressionKind::Unary: return this -> EvaluateBooleanUnary((BoundUnaryExpression*)expression);
        case BoundExpressionKind::Binary: return this -> EvaluateBooleanBinary((BoundBinaryExpression*)expression);
        case BoundExpressionKind::Variable: return dynamic_pointer_cast<BooleanEvalValue>(this->GetVariable((BoundVariableExpression*)expression));
        case BoundExpressionKind ::FunctionCall: return dynamic_pointer_cast<BooleanEvalValue>(this->EvaluateFunctionCallExpression(expression));

        case BoundExpressionKind::Bad:
        case BoundExpressionKind::LiteralInteger:
        case BoundExpressionKind::LiteralFloat:
        case BoundExpressionKind::LiteralString:
            throw;

    }
}

shared_ptr<StringEvalValue> Evaluator::EvaluateStringExpression(BoundExpression *expression) {
    switch (expression->GetKind()) {
        case BoundExpressionKind ::LiteralString:
            return make_shared<StringEvalValue>(((BoundLiteralStringExpression*)expression)->GetValue());
        case BoundExpressionKind::Binary:
            return this -> EvaluateStringBinary((BoundBinaryExpression*)expression);
        case BoundExpressionKind::Variable: return dynamic_pointer_cast<StringEvalValue>(this->GetVariable((BoundVariableExpression*)expression));
        case BoundExpressionKind ::FunctionCall: return dynamic_pointer_cast<StringEvalValue>(this->EvaluateFunctionCallExpression(expression));

        case BoundExpressionKind::Bad:
        case BoundExpressionKind::LiteralInteger:
        case BoundExpressionKind::LiteralFloat:
        case BoundExpressionKind::LiteralBool:
        case BoundExpressionKind::Unary:
            throw;

    }
}

shared_ptr<EvalValue> Evaluator::EvaluateFunctionExpression(BoundExpression * expression){
    switch (expression->GetKind()){
        case BoundExpressionKind ::Variable: return this->GetVariable((BoundVariableExpression*)expression);
        default: throw;
    }
}
shared_ptr<EvalValue> Evaluator::EvaluateNilExpression(BoundExpression * expression){
    switch (expression->GetKind()){
        case BoundExpressionKind ::FunctionCall:
            return this->EvaluateFunctionCallExpression(expression);
        default:
            return nullptr;
    }
}


shared_ptr<EvalValue> Evaluator::EvaluateFunctionCallExpression(BoundExpression* expression){
    auto functionCall = (BoundFunctionCallExpression*)expression;
    auto function = dynamic_pointer_cast<ScriptFunctionEvalValue>(this->EvaluateExpression(functionCall->GetFunctionExpression()));
    auto boundParameters = functionCall->GetParameters();
    auto parameters = vector<shared_ptr<EvalValue>>(boundParameters.size());
    for (int i = 0; i < boundParameters.size(); i++){
        parameters[i] = this->EvaluateExpression(boundParameters[i]);
    }

    auto type = std::dynamic_pointer_cast<FunctionScriptType>(function->GetType());
    auto parameterTypes = type->GetParameterTypes();
    auto parameterKeys = type->GetParameterKeys();
    for (int i = 0; i < parameterTypes->size() && i < parameterKeys->size() && i < parameters.size(); i++){
        auto parameter = parameters[i];
        auto requiredType = parameterTypes->at(i);
        if (*parameter->GetType() != requiredType.get()){
            throw EvaluationException("Passed wrong type to function.");
        }
        auto key = parameterKeys->at(i);
        this->_evaluationScope->CreateVariable(key->GetScopeId(), key->GetIdentifier(), parameter->Clone());
    }
    this->EvaluateBlockStatement(function->GetInnerBlock().get());
    return nullptr;

}

